/* Dawicontrol generic Linux driver, 18.08.2023

   Copyright (C) 2022-2023 Dawicontrol Computersysteme GmbH

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License version 2 as published 
   by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along 
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "dcmodule.h"


/****************************************************************************/
/*      Int 13h compatible CHS-translation                                  */
/****************************************************************************/

static int chs_xlat(struct scsi_device *sdev, struct block_device *bdev,
                    sector_t capacity, int chs[])
{
    // use the controller-BIOS translation-scheme
    chs[0]  = 255;  // Heads
    chs[1]  = 63;   // Sectors
    chs[2]  = sector_div(capacity, 255*63);
    return 0;
}

/****************************************************************************/
/*      Command handler                                                     */
/****************************************************************************/

static int queue_cmd(struct Scsi_Host *host, struct scsi_cmnd *cmdblk)
{
    struct SCSI_req_blk *srb;
    int i;

    // log command parameters
    #ifdef debug_mode
        int bus, id, lun;
        bus = cmdblk->device->channel;
        id = cmdblk->device->id;
        lun = cmdblk->device->lun;
        printk(KERN_DEBUG my_name ": ********************************************************************\n");
        printk(KERN_DEBUG "Bus = %d,  Target = %d,  Lun = %d\n", bus, id, lun);
        printk(KERN_DEBUG "CDB = 0x");
        for (i = 0; i < cmdblk->cmd_len; ++i)
            printk(KERN_CONT "%.2x ", cmdblk->cmnd[i]);
        printk(KERN_DEBUG "scsi_sg_count = %d,  scsi_bufflen = %d,  scsi_sglist = 0x%px\n",
               scsi_sg_count(cmdblk), scsi_bufflen(cmdblk), scsi_sglist(cmdblk));
    #endif

    // get SRB
    srb = scsi_cmd_priv(cmdblk);
    srb->len = sizeof (struct SCSI_req_blk);
    srb->ver = srb_version;
    srb->ref = cmdblk;

    // get CDB
    i = cmdblk->cmd_len;
    srb->cdblen = i;
    memcpy(srb->cdb, cmdblk->cmnd, i);

    // get bus/target/lun
    srb->bus    = cmdblk->device->channel;
    srb->target = cmdblk->device->id;
    srb->lun    = cmdblk->device->lun;

    // get direction
    switch(cmdblk->sc_data_direction) {
        case DMA_FROM_DEVICE:
            srb->flags = srb_flag_dir_in; 
            break;
        case DMA_TO_DEVICE:
            srb->flags = srb_flag_dir_out;
            break;
        case DMA_BIDIRECTIONAL:
            srb->flags = srb_flag_dir_any;
            break;
        default:
            srb->flags = srb_flag_dir_none;
            break;
    }
    // get transfer parameters
    srb->sensebuffer= cmdblk->sense_buffer;
    srb->sg_list    = scsi_sglist(cmdblk);
    srb->xfersize   = scsi_bufflen(cmdblk);
    srb->sg_count   = scsi_sg_count(cmdblk);

    // get buffer mapping
    if (scsi_dma_map(cmdblk) < 0) {
        return SCSI_MLQUEUE_HOST_BUSY;
    }
    // log mapped SG parameters
    #ifdef debug_mode
        // walk_sg(srb->sg_list);
        printk(KERN_DEBUG "****************************************************************************\n");
    #endif

    // pass SRB & return
    return BIN_exec_req(srb, &host->hostdata);
}

/****************************************************************************/
/*      Error handler callback functions                                    */
/****************************************************************************/

//***** Bus reset

static int bus_reset(struct scsi_cmnd *cmdblk)
{
    int bus;
    struct Scsi_Host *host;
    
    #ifdef debug_mode
        printk(KERN_ERR my_name ": Bus Reset\n");
    #endif

    if (!cmdblk) {
        return FAILED;
    }
    host = cmdblk->device->host;
    bus = cmdblk->device->channel;
    return BIN_bus_reset(&host->hostdata, bus);
}

//***** Host reset

static int host_reset(struct scsi_cmnd *cmdblk)
{
    struct Scsi_Host *host;

    #ifdef debug_mode
        printk(KERN_ERR my_name ": Host Reset\n");
    #endif

    if (!cmdblk) {
        return FAILED;
    }
    host = cmdblk->device->host;
    return BIN_hba_reset(&host->hostdata);
}

/****************************************************************************/
/*      PROC file functions                                                 */
/****************************************************************************/

//***** get status of controllers, drives and RAID sets

static int show_procfile(struct seq_file *m, void *v)
{
    char buffer[80];
    int hn, dn, rn;

    seq_puts(m, my_description);
    seq_puts(m, " Vers ");
    seq_puts(m, my_version);
    seq_puts(m, "\n\n");

    // list all controllers managed by this driver
    hn = 0;
    if (!BIN_get_hba(&buffer, &hn)) {
        seq_puts(m, "No controllers found.\n");
    } else {
        while (BIN_get_hba(&buffer, &hn)) {
            seq_puts(m,"\n");
            seq_puts(m, "Controller   BIOS   Irq   PCI address    Busmode\n");
            seq_puts(m, "-------------------------------------------------------------\n");
            seq_puts(m, buffer);
            seq_puts(m, "\n\n");

            // list all drives attached to this controller
            dn=0;
            if (!BIN_get_device(&buffer, &hn, &dn)) {
                seq_puts(m, "No disk drives found.\n");
            } else {
                seq_puts(m, "Drive     Model                          Size     Status  Set\n");
                seq_puts(m, "-------------------------------------------------------------\n");
                while (BIN_get_device(&buffer, &hn, &dn)) {
                    seq_puts(m, buffer);
                    seq_puts(m, "\n");
                    dn++;
                };
                seq_puts(m, "\n");

                // list all RAID sets configured by these drives
                rn=0;
                if (!BIN_get_raid(&buffer, &hn, &rn)) {
                    seq_puts(m, "No RAID sets found.\n");
                } else {
                    seq_puts(m, "RAID Set  Type       Blocksize  Members  Size     Status\n");
                    seq_puts(m, "-------------------------------------------------------------\n");
                    while (BIN_get_raid(&buffer, &hn, &rn)) {
                        seq_puts(m, buffer);
                        seq_puts(m, "\n");
                        rn++;
                    };
                };
            };
            seq_puts(m, "\n");
            hn++;
        };
    };
    seq_puts(m, "\n");
    return 0;
}

//***** open PROC file

static int open_procfile(struct inode *inode, struct file *file)
{
    return single_open(file, show_procfile, NULL);
}

//***** PROC file operations

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
    static const struct proc_ops my_proc_ops =
    {
        .proc_open      = open_procfile,
        .proc_read      = seq_read,
        .proc_lseek     = seq_lseek,
        .proc_release   = single_release,
    };
#else
    static const struct file_operations my_proc_ops =
    {
        .owner      = THIS_MODULE,
        .open       = open_procfile,
        .read       = seq_read,
        .llseek     = seq_lseek,
        .release    = single_release,
    };
#endif


/****************************************************************************/
/*      PCI device probe function                                           */
/****************************************************************************/

static int hba_size = 0;
static struct scsi_host_template host_template = 
{
    .module                 = THIS_MODULE,
    .name                   = my_name,
    .cmd_size               = sizeof (struct SCSI_req_blk),
    .bios_param             = chs_xlat,
    .queuecommand           = queue_cmd,
    .eh_bus_reset_handler   = bus_reset,
    .eh_host_reset_handler  = host_reset,
    .proc_name              = my_name,
    .this_id                = -1,
    .can_queue              = 32,
    .cmd_per_lun            = 32,
    .sg_tablesize           = 0,
    .max_sectors            = 0,
    .max_segment_size       = my_max_segment_size,
    .dma_boundary           = my_dma_boundary,
    .emulated               = 0,
};

static int probe_device(struct pci_dev *pdev, const struct pci_device_id *id)
{   
    int error;
    struct Scsi_Host *host;

    // alloc host struc with hba extension
    host = scsi_host_alloc(&host_template, hba_size);
    if (!host) {
        printk(KERN_ERR my_name ": cannot allocate host.\n");
        return -ENOMEM;
    }
    // pass pci_dev, scsi_host & hba ptr
    // and initialize controller
    error = BIN_init_hba(pdev, host, &host->hostdata);
    if (error) {
        printk(KERN_ERR my_name ": cannot initialize host.\n");
        scsi_host_put(host);
        return error;
    }
    // setup host parameter
    host->max_channel = my_number_of_ports-1;
    host->max_id = 1;
    host->max_lun = 2;
    host->max_cmd_len = max_cdblen;
    host->nr_hw_queues = my_number_of_ports;

    // register to upper layer
    error = (scsi_add_host(host, &pdev->dev));
    if (error) {
        printk(KERN_ERR my_name ": cannot register host.\n");
        scsi_remove_host(host);
        BIN_remove_hba(pdev, &host->hostdata);
        scsi_host_put(host);
        return error;
    }
    // link pci_dev to host and collect drives
    pci_set_drvdata(pdev, host);
    scsi_scan_host(host);
    return 0;
}

/****************************************************************************/
/*      PCI device remove function                                          */
/****************************************************************************/

static void remove_device(struct pci_dev *pdev)
{
    struct Scsi_Host *host;

    // get host and remove
    host = pci_get_drvdata(pdev);
    if (host) {
        scsi_remove_host(host);
        BIN_remove_hba(pdev, &host->hostdata);
        scsi_host_put(host);
    }
    return;
}

/****************************************************************************/
/*      PCI device suspend function                                         */
/****************************************************************************/

static int suspend_device(struct pci_dev *pdev, pm_message_t state)
{
    struct Scsi_Host *host;

    #ifdef debug_mode
        printk(KERN_DEBUG my_name ": suspend host\n");
    #endif

    // get host and suspend
    host = pci_get_drvdata(pdev);
    if (host) {
        BIN_suspend_hba(pdev, &host->hostdata);
    }
    return 0;
}

/****************************************************************************/
/*      PCI device resume function                                          */
/****************************************************************************/

static int resume_device(struct pci_dev *pdev)
{
    struct Scsi_Host *host;

    #ifdef debug_mode
        printk(KERN_DEBUG my_name ": resume host\n");
    #endif

    // get host and resume
    host = pci_get_drvdata(pdev);
    if (host) {
        BIN_resume_hba(pdev, &host->hostdata);  
    }
    return 0;
}

/****************************************************************************/
/*      PCI device shutdown function                                        */
/****************************************************************************/

static void shutdown_device(struct pci_dev *pdev)
{
    struct Scsi_Host *host;
    int shutdown = 0;

    // get host and shutdown
    host = pci_get_drvdata(pdev);
    if (host) {
        // check if shutdown or restart
        if (system_state == SYSTEM_POWER_OFF) {
            shutdown = 1;
        }
        BIN_shutdown_hba(pdev, &host->hostdata, shutdown);
    }
    return;
}

/****************************************************************************/
/*      Install / uninstall driver module                                   */
/****************************************************************************/

static struct pci_driver my_pci_drv =
{
    .name       = my_name,
    .id_table   = my_pci_ids,
    .probe      = probe_device,
    .remove     = remove_device,
    .suspend    = suspend_device,
    .resume     = resume_device,
    .shutdown   = shutdown_device,
};

static int __init install_driver(void)
{
    int error;
    
    // get hba_size and number of sg-entrys
    hba_size = BIN_init_driver(sizeof(struct timer_list), 
	    &host_template.sg_tablesize, GFP_KERNEL, GFP_ATOMIC);
    if (!hba_size) {
        return -ENODEV;
    }
    if (!host_template.sg_tablesize) {
        return -ENODEV;
    }
    host_template.max_sectors = host_template.sg_tablesize * 8;
    
    // setup proc
    proc_create(my_name, 0, NULL, &my_proc_ops);
    
    // initialize
    error = pci_register_driver(&my_pci_drv);
    if (error) {
        printk(KERN_ERR my_name ": cannot register driver.\n");
    } else {
        printk(KERN_NOTICE my_name ": RAID driver installed.\n");
    }
    return error;
}

static void __exit uninstall_driver(void)
{
    pci_unregister_driver(&my_pci_drv);
    remove_proc_entry(my_name, NULL);
    printk(KERN_NOTICE my_name ": RAID driver removed.\n");
    return;
}

/****************************************************************************/
/*      Module description                                                  */
/****************************************************************************/

MODULE_VERSION(my_version);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(my_description);
MODULE_AUTHOR("Dawicontrol Computersysteme GmbH");

module_init(install_driver);
module_exit(uninstall_driver);
